/*
...
*/

// in case we use Eigen
#ifdef USE_EIGEN_LU_IMPLEMENTATION

#ifndef MATRIX_EIGEN_H
#define MATRIX_EIGEN_H

#include "vector_2.h"

#include <Eigen/Core>

namespace Eigen {
    template <class MatrixType> class LU;
}

#include <ostream>


    // this should be removed one day...
#define CUI const unsigned int

class Matrix {
    public:
        /**
        * Creates a size x size square matrix m, with all values zero,
        * and a right side vector x of size m+n
         */
        Matrix ( CUI size );

        /**
         * Destructor
         */
        ~Matrix();

        /**
         * Returns true if the matrix is changed since last calling performLU()
         * - i.e. if we do need to call performLU again.
         */
        bool isChanged() const { return m_isChanged; }

        /**
         * Performs LU decomposition. Going along the rows,
         * the value of the decomposed LU matrix depends only on
         * the previous values.
         */
        void performLU();
        /**
         * Applies the right side vector (x) to the decomposed matrix,
         * with the solution returned in x.
         */
        void fbSub ( QuickVector *x );

        /**
         * Prints the matrix to an output stream
         * @param outstream the stream to print to
         */
        void displayMatrix ( std::ostream &outstream ) const;

        /**
         * Prints the LU-decomposed matrix to an output stream
         * @param outstream the stream to print to
         */
        void displayLU ( std::ostream &outstream ) const;

        /**
         * Set the matrix to identity
         */
        // void identity();

        /**
         * get / set elements in the matrix
         * @param i
         * @param j
         * @return a reference to the value at given position: can be read/write
         *
         * Here is a problem: we don't know when is the value read and when it's written
         * So we can't optimize for skipping LU
         */
        double &g ( CUI i, CUI j );


        /**
         * Multiplies this matrix by the Vector rhs, and places the result
         * in the vector pointed to by result. Will fail if wrong size.
         * @param rhs vector to multiply the vector with
         * @param result matrix * rhs, vector @todo check if the order is right
         */
        void multiply ( const QuickVector *rhs, QuickVector *result );

        /**
         * \fn Matrix::validateLU()
         * check the validity of LU factorization.
         * It's an expensive procedure only for debugging.
         * @return sum of absolute error between the original matrix and the decomposition
         *
         * CAUTION: SUSPECT BUGS IN THIS FUNCTION JUST AS QUICKLY AS BUGS IN THE ABOVE.
         */
        // double validateLU() const;

        /**
         * @fn Matrix::validate()
         * If this here function fails, the matrix is no good and
         * no further computation should be done. This method should be used before performLU.
         * @return true, if the matrix is valid, false otherwise
         */
        // bool validate() const;

    protected:
        // TODO add pointers so compilation of this header will be considerable faster
        Eigen::MatrixXd *m_matr; /// input matrix

        Eigen::LU<Eigen::MatrixXd> *m_lu;  /// the LU decomposition of a matrix

        bool m_isChanged; /// set to true when the matrix had changed since the last LU decomposition
};

// these come in matrix_eigen.cpp, but lazyness rulez

#include "Eigen/Core"
#include "Eigen/LU"
#include "vector_2.h"
                 
Matrix::Matrix ( CUI size ) {
    m_isChanged = true;
    m_lu = NULL;
    m_matr = new Eigen::MatrixXd ( ( int ) size, ( int ) size );
}

Matrix::~Matrix() {
    delete m_matr;
    if ( m_lu != NULL )
        delete m_lu;
}

void Matrix::performLU() {
    //if( m_isChanged ){
    // don't leak memory
    if ( m_lu != NULL )
        delete m_lu;
    // create new LU decomposition
    m_lu = new Eigen::LU<Eigen::MatrixXd> ( *m_matr );
    //} else
    // make sure m_lu != null
    //  if( m_lu == NULL)
    m_lu = new Eigen::LU<Eigen::MatrixXd> ( *m_matr );
}

void Matrix::multiply ( const QuickVector *rhs, QuickVector *result ) {
    // here are quite a few indirections...
    Eigen::VectorXd t = ( *m_matr ) * ( *rhs );
    (*result) = t; // ( *m_matr ) * ( *rhs );
}

void Matrix::fbSub ( QuickVector *x ) {
    m_lu->solve ( *x, x ); // great...
}

double & Matrix::g ( CUI i, CUI j ) {
    return ( *m_matr ) ( i,j );
}

void Matrix::displayMatrix ( std::ostream &outstream ) const {
    outstream << (*m_matr);
}

void Matrix::displayLU ( std::ostream &outstream ) const {
    outstream << m_lu->matrixLU();
}


#endif // MATRIX_EIGEN_H

#endif // USE_EIGEN_LU_IMPLEMENTATION
